-- [[ Namespaces ]] --
local _, addon = ...;
local diagnostics = addon.Diagnostics;
local objects = addon.Objects;
objects.Achievement = {};
local achievement = objects.Achievement;

-- [[ Constructors ]] --
achievement.__index = achievement;
function achievement:New(id, points, faction, covenant, obtainable, hasWowheadLink) -- Creates a new achievement
    local self = {};
    setmetatable(self, achievement);

    self.ID = id or 0;
    self.Points = points or 0;
    self.Faction = faction;
    self.Covenant = covenant;
    if obtainable == false then -- We only want to set it if it's not obtainable, otherwise nil
        self.NotObtainable = true; -- By inverting this we reduce memory usage because most are obtainable
    end
    if hasWowheadLink == false then -- We only want to set it if it has no Wowhead link, otherwise nil
        self.HasNoWowheadLink = true; -- By inverting this we reduce memory usage because most have a Wowhead link
    end

    return self;
end

function achievement:GetMergedCategory() -- Gets the achievement's category, used when merging categories is enabled
    diagnostics.Trace("achievement:GetCategory");

    local categories = self.Category:GetTree(); -- Issue #43: Fix
    for _, category in next, categories do
        if category.MergedAchievements ~= nil then
            for _, ach in next, category.MergedAchievements do
                if ach.ID == self.ID then
                    -- diagnostics.Debug(category.Name);
                    return category;
                end
            end
        end
        if category.Achievements ~= nil then
            for _, ach in next, category.Achievements do
                if ach.ID == self.ID then
                    -- diagnostics.Debug(category.Name);
                    return category;
                end
            end
        end
    end

    error("The achievement with ID " .. self.ID .. " has no category."); -- Should in theory never happen, means there is a problem with the database
end

function achievement:GetRequiredForIDs(validate, filters)
    local criteriaCache = addon.BuildCache();

    if self.RequiredForIDs then -- Return cached list
        return self.RequiredForIDs;
    end

    self.RequiredForIDs = {};
	for _, criteria in next, criteriaCache do
		if criteria.AchievementID == self.ID then
            if validate and filters and addon.Data.Achievements[criteria.RequiredForID] then
                if validate(filters, addon.Data.Achievements[criteria.RequiredForID], true) > 0 then
			        tinsert(self.RequiredForIDs, criteria.RequiredForID);
                end
            else
			    tinsert(self.RequiredForIDs, criteria.RequiredForID);
            end
		end
	end

    if #self.RequiredForIDs == 0 then
        self.RequiredForIDs = nil;
        return;
    end

    return self.RequiredForIDs;
end

function achievement:GetPartOfAChainIDs(validate, filters)
    if self.PartOfAChainIDs then -- Return cached list
        return self.PartOfAChainIDs;
    end

    if self.ID == 15077 or self.ID == 15078 then -- Issue #49 : Fix
        return;
    end

	if not GetNextAchievement(self.ID) and not GetPreviousAchievement(self.ID) then
		return;
	end

	local id;
	local tmpID = self.ID;
	while tmpID do -- Find first achievement in a chain
		id = tmpID;
		tmpID = GetPreviousAchievement(id);
	end

    self.PartOfAChainIDs = {}; -- By creating the part of a chain IDs table here we reduce memory usage because not every achievement has part of a chain IDs

	while id do
        if validate and filters and addon.Data.Achievements[id] then
            if validate(filters, addon.Data.Achievements[id], true) > 0 then
                tinsert(self.PartOfAChainIDs, id);
            end
        else
            tinsert(self.PartOfAChainIDs, id);
        end
		id = GetNextAchievement(id);
	end

    return self.PartOfAChainIDs;
end

function achievement:AddCategory(category)
    if self.Category == nil then
        self.Category = category;
    else
        if self.MoreCategories == nil then
            self.MoreCategories = {};
        end
        tinsert(self.MoreCategories, category);
    end
end

function achievement:Include()
    self.Excluded = nil;
    if SavedData.ExcludedAchievements == nil then
        return;
    end

    SavedData.ExcludedAchievements[self.ID] = nil;
end

function achievement:Exclude()
    self.Excluded = true;
    if SavedData.ExcludedAchievements == nil then
        SavedData.ExcludedAchievements = {};
    end

    SavedData.ExcludedAchievements[self.ID] = true;
end

function achievement:ClearFocus()
    self.Focused = nil;
    if SavedData.FocusedAchievements == nil then
        return;
    end

    SavedData.FocusedAchievements[self.ID] = nil;
end

function achievement:Focus()
    self.Focused = true;
    if SavedData.FocusedAchievements == nil then
        SavedData.FocusedAchievements = {};
    end

    SavedData.FocusedAchievements[self.ID] = true;
end